
--[[

name:   Antix LogWindow
vers:   1.0.2
desc:   A simple log window that you can append text to.
auth:   by Cliff Earl
date:   June 2016
legal:  Distribute/ modify as you like, but retain my credits, thanks

histo:  v1.0.1 - No longer subclassed from shape
               - Tidied up init/dispose code
		v1.0.2 - No longer splits into p[aragraphs, just words
			   - Uses /n to create new lines
		v1.0.3 - Stores lines in a buffer that can be drag scrolled through
--]]

LogWindow = Core.class(Sprite)

require "bit"

function LogWindow:init (options)
  local x, y, textColor, scale
  
  local props = { -- Default options
    lines           = 12,
    font            = Font.getDefault(),
    x               = 16,
    y               = 16,
    width           = 300,
    textColor       = 0xeeeef0,
    scale           = 1,
    width           = 512,
	dragEnabled		= true,
	enabled			= true,
  }
  
  if options then -- overlay any specified options
    for key, value in pairs(options) do
      props[key]= value
    end
  end

  self.lines        = props.lines
  self.font         = props.font
  x                 = props.x
  y                 = props.y
  self.width        = props.width
  self.textColor    = props.textColor
  scale             = props.scale
  self.dragEnabled  = props.dragEnabled
  self.enabled      = props.enabled
  self.topLine      = 0

  self:setScale(scale)
  self:setPosition(x, y)

  local lineHeight = self.font:getLineHeight()
  
  local textLines = {} -- Create blank text lines
  for i = 1, self.lines do
    textLines[#textLines + 1] = {r = 0.9, g = 0.9, b = 0.95, line = " "}
    local text = TextField.new(self.font, " ")
    text:setPosition(8, (i * lineHeight) + 8)
    text:setTextColor(self.textColor)
    self:addChild(text)
  end
  self.textLines = textLines
  self.lineHeight = lineHeight

  self:addEventListener(Event.TOUCHES_BEGIN, self.pressed, self) -- Add event handlers
  self:addEventListener(Event.TOUCHES_MOVE, self.dragged, self)
  self:addEventListener(Event.TOUCHES_END, self.released, self)
end

-- Return the number of buffered lines.
function LogWindow:count()
  return #self.textLines
end

-- Handle touch event
function LogWindow:pressed(e)
  if not self.enabled then return end

  if self:hitTestPoint(e.touch.x, e.touch.y) then
    e:stopPropagation()
    self.id = e.touch.id
    self.focus = true
    self.handleY = e.touch.y
  end
end

-- Handle drag event
function LogWindow:dragged(e)
  if not self.enabled then return end

  local function clamp(n, min, max)
    return n < min and min or (n > max and max or n)
  end

  if e.touch.id == self.id then
    self.focus = true
    e:stopPropagation()
    if self.dragEnabled then
	  local newScroll = self.topLine
      local dy = e.touch.y - self.handleY
      local maxScroll = (#self.textLines - self.lines - 1) * self.lineHeight
      local row = clamp(newScroll - dy, 0, maxScroll)
	  self.topLine = row
      row = math.floor(row / self.lineHeight)  + 1
	  self:redraw(row)
      self.handleY = e.touch.y
    end
   end
end

-- Handle released event
function LogWindow:released(e)
  if not self.enabled then return end

  if e.touch.id == self.id and self.focus and self:hitTestPoint(e.touch.x, e.touch.y) then
    e:stopPropagation()
    self.focus = false
    self.id = -1
  end
end

-- Attempt to cleanup our mess :D
function LogWindow:shutDown()
  self.enabled = false

  self:removeEventListener(Event.TOUCHES_BEGIN, self.pressed, self) -- Remove event handlers
  self:removeEventListener(Event.TOUCHES_MOVE, self.dragged, self)
  self:removeEventListener(Event.TOUCHES_END, self.released, self)

  self:empty(self.textLines)
  for i = self:getNumChildren(), 1, -1 do -- Remove in reverse order
    local text = self:getChildAt(i)
    self:removeChildAt(i)
    text = nil
  end
  self.lines = nil
  self.font = nil
  self.width = nil
  self.textColor = nil
  self.textLines = nil
end

-- Clear all lines of text and empty text buffer
function LogWindow:clear()
  local textLines = self.textLines
  self:empty(textLines)
  for i = 1, self:getNumChildren() do -- Create new textlines
    textLines[#textLines + 1] = {r = 0.9, g = 0.9, b = 0.95, line = ""}
    self:getChildAt(i):setText("")
  end
end

-- Empty a table, and all tables contained within it
function LogWindow:empty(table)
  for k in next, table do
    local t = type(k)
    if k == "table" then -- If next is a table, empty that
      self:empty(k)
    end
    table[k] = nil
  end
end

-- Attach LogWindow to new parent
function LogWindow:attach(newParent)
  self:detach()
  newParent:addChild(self)
end

-- Detach LogWindow from current parent
function LogWindow:detach()
  self:removeFromParent()
end

-- Show/Hide
function LogWindow:show(show)
  self:setVisible(show)
end

-- Bring to front of all others sprites attached to parent
function LogWindow:toFront()
  self:getParent():addChild(self)
  self:setVisible(true)
end

-- Append new text to end of log
function LogWindow:append(text, color)
  local r, g, b = self:colorToRGB(color or self.textColor)
  
  self:split(r, g, b, text, self.width - 16)

  local row = #self.textLines - self.lines
  if row < 1 then row = 1 end
  self.topLine = row * self.lineHeight
  
  self:redraw(#self.textLines - self.lines)
end

-- Redraw the log from any line in the buffer
function LogWindow:redraw(row)

  local textLines = self.textLines
  local visibleRows = self.lines

  if row < 1 then -- Determine top row
    row = 1
  elseif row > visibleRows then
    row = #textLines - visibleRows
  end

  for i = 1, visibleRows do -- Redraw rows
    local line = textLines[i + row]
    local text = self:getChildAt(i)
    text:setText(line.line)
    text:setColorTransform(line.r, line.g, line.b)
  end
end

-- Split text into multiple lines
function LogWindow:split(r, g, b, text, maxWidth)

	local function split(input, char) -- SPLIT FUNCTION FROM STACK EXCHANGE http://bit.ly/1iXGtNN
		local match = string.gmatch
		local output = {}
		local pattern = '([^'..char..']+)' -- OPTIMIZED SLIGHTLY BY ANTIX
		for str in match(input, pattern) do -- MORE OPTIMIZATIONS AFTER FULLY READING THREAD
			output[#output + 1] = str -- THIS IS FASTER THAN TABLE.INSERT
		end
		return output
	end

	local words = split(text, " ") -- Split text into words

	local lines = self.textLines

	local line = ""
	local lineWidth = 0

	local font = self.font

	local topLine = "/n" -- new line identifier
	local FAKEtopLine = "//n"

	for i = 1, #words do
		local word = words[i] -- Next word to process
		
		if word == topLine then -- Check for and insert new line if required
			lines[#lines + 1] = {r = r, g = g, b = b, line = ""}
				line = ""
				lineWidth = 0
			else
				if word == FAKEtopLine then -- Sometimes we might actually want top print /n
					word = topLine
			end
			local wordWidth = font:getAdvanceX(word.." ") -- Get pixel width of next word
			if lineWidth + wordWidth >= maxWidth then -- Create a new line if the width is exceeded
				lines[#lines + 1] = {r = r, g = g, b = b, line = line}
				line = ""
				lineWidth = 0
			end
			line = line .. " " .. word -- Add word to line (along with SPACE that was stripped during split)
			lineWidth = lineWidth + wordWidth 
		end
	end
	lines[#lines + 1] = {r = r, g = g, b = b, line = line} -- Always append the final line
end

-- Convert RGB to R,G,B saturation
function LogWindow:colorToRGB(color)
  local r = 1/255 * bit.band(bit.ror(color, 16), 255)
  local g = 1/255 * bit.band(bit.ror(color, 8), 255)
  local b = 1/255 * bit.band(color, 255)
  return r, g, b
end
